/* $Id: CrowdAuthSecurityFilters.groovy 41035 2008-05-09 05:59:08Z caulfield14 $ */

import org.codehaus.groovy.grails.commons.ApplicationHolder
import org.codehaus.groovy.grails.commons.GrailsClassUtils
import org.codehaus.groovy.grails.commons.ConfigurationHolder as CH
import com.atlassian.crowd.integration.http.HttpAuthenticator
import com.atlassian.crowd.integration.service.soap.client.SecurityServerClient
import org.codehaus.grails.plugins.crowd.util.CrowdAuthUtils


/**
 * Filters that provide authentication and authorization for requests
 */
class CrowdAuthSecurityFilters {
    HttpAuthenticator httpAuthenticator
    SecurityServerClient securityServerClient

    def filters = {
        log.info("Initializing Crowd Auth Security Filters")

        /**
         * Runs on every controller and action request. Determines if the calling controller
         * and/or action require authentication, declared globally in a configuration variable
         * called "defaultAuthenticate" (default: false) that can be overridden by a static
         * variable on the controller called "authenticate". If the request needs
         * authentication, the filter will check to see if the user is already logged in via
         * Crowd's SSO. If so, the user will be automatically logged in to the application.
         * If not, the user will be redirected to a URL that can log the user in (default:
         * LoginController). If the user is logged in via Crowd's SSO but not the application,
         * the user will be logged in regardless of whether the request requires
         * authentication or not.
         */
        authenticationCheck(controller: '*', action: '*') {
            before = {
                // Note: controllerName and actionName are given to us by Spring/Grails

                def application= ApplicationHolder.application
                def applicationControllerClasses= application.controllerClasses

                def controllerClass = applicationControllerClasses.find() { it.logicalPropertyName == controllerName }
                if (!controllerClass) {
                    log.warn("Couldn't find a Controller class associated with '${controllerName}'")
                    return true;
                }
                else log.debug("Controller class '${controllerClass}' located")

                def requireAuthentication= false

                // Process default authenticate
                def defaultAuthenticate= CH.config.crowdAuth.defaultAuthenticate
                log.debug("DefaultAuthenticate: ${defaultAuthenticate}")
                switch (defaultAuthenticate) {
                case Boolean:
                    requireAuthentication= defaultAuthenticate
                    break
                case Map:
                    def except= defaultAuthenticate['except']
                    def only= defaultAuthenticate['only']

                    if (except) {
                        if ((actionName? actionName : 'index') in except)
                            requireAuthentication= false
                        else
                            requireAuthentication= true
                    }
                    else if (only) {
                        if ((actionName? actionName : 'index') in only)
                            requireAuthentication= true
                        else
                            requireAuthentication= false
                    }
                    else {
                        log.warn("DefaultAuthenticate was not correctly defined in configuration.")
                    }
                    break
                default:
                    log.warn("DefaultAuthenticate was not correctly defined in configuration.")
                }

                // Process Controller/Action supplied authenticate
                Object authenticate= GrailsClassUtils.getStaticPropertyValue(controllerClass.getClazz(), "authenticate");
                switch (authenticate) {
                case Boolean:
                    requireAuthentication= authenticate
                    break
                case Map:
                    def except= authenticate['except']
                    def only= authenticate['only']

                    if (except) {
                        if ((actionName? actionName : 'index') in except)
                            requireAuthentication= false
                        else
                            requireAuthentication= true
                    }
                    else if (only) {
                        if ((actionName? actionName : 'index') in only)
                            requireAuthentication= true
                        else
                            requireAuthentication= false
                    }
                    else {
                        log.warn("Authenticate was not correctly defined in Controller '${controllerClass}'")
                    }
                    break
                default:
                    log.warn("Authenticate was not correctly defined in Controller '${controllerClass}'")
                }

                // Automatically authenticate a user a the application level if the user has Crowd SSO authentication
                // Even if the request does not require authentication
                if (!CrowdAuthUtils.isAuthenticated(request) && httpAuthenticator.isAuthenticated(request, response)) {
                    // TODO: Catch Exceptions?
                    session.principal= httpAuthenticator.getPrincipal(request)
                    session.principalGroups= securityServerClient.findGroupMemberships(session.principal.name).toList()
                    session.principalRoles= securityServerClient.findRoleMemberships(session.principal.name).toList()
                }

                // Is user's application level authentication valid?
                if (!CrowdAuthUtils.isAuthenticated(request)) {
                    log.debug("User is not authenticated at application level")

                    // Does the user have Crowd SSO authentication?
                    if (httpAuthenticator.isAuthenticated(request, response)) {
                        // The user has Crowd SSO authentication
                        // TODO: Catch Exceptions?
                        session.principal= httpAuthenticator.getPrincipal(request)
                        session.principalGroups= securityServerClient.findGroupMemberships(session.principal.name).toList()
                        session.principalRoles= securityServerClient.findRoleMemberships(session.principal.name).toList()
                    }
                    else {
                        // The user does not have Crowd SSO authentication
                        log.debug("User is not authenticated at Crowd SSO level")

                        // If the request requires authentication, store original request and redirect to login
                        if (requireAuthentication) {
                            def originalRequestParams = [controller: controllerName, action: actionName? actionName : 'index']
                            originalRequestParams.putAll(params)
                            session.originalRequestParams = originalRequestParams

                            // User is not authenticated, redirect the user to login
                            if (CH.config.crowdAuth.loginAction) {
                                log.debug("Redirecting to loginController, loginAction")
                                redirect(controller: CH.config.crowdAuth.loginController, action: CH.config.crowdauth.loginAction)
                                return
                            }
                            else if (CH.config.crowdAuth.loginController) {
                                log.debug("Redirecting to loginController")
                                redirect(controller: CH.config.crowdAuth.loginController)
                                return
                            }
                            else if (CH.config.crowdAuth.loginUrl) {
                                log.debug("Redirecting to loginUrl")
                                redirect(uri: CH.config.crowdAuth.loginUrl)
                                return
                            }
                            else {
                                log.error("No login method is configured. Please set one of loginController [and loginAction] or loginUrl.")
                                return false
                            }
                        }
                    }
                }
            }
        }

        /**
         * Runs on every controller and action request. Determines if the calling controller
         * and/or action require authorization, declared globally in a configuration variable
         * called "defaultAuthorize" (default: none) that can be overridden by a static
         * variable on the controller called "authorize". If the request needs authorization,
         * the filter gets a list of groups and roles that the user belongs to and determines
         * if the user meets the authorization requirements for the controller and/or action.
         * If the user is not authorized, an HTTP 503 FORBIDDEN response is returned. (The
         * response can be customized via Grails UrlMappings.)
         *
         * TODO: Currently, only Crowd Groups are supported for authorization. Role-based
         * authorization needs to be implemented.
         */
        authorizationCheck(controller: '*', action: '*') {
            before = {
                def application= ApplicationHolder.application
                def applicationControllerClasses= application.controllerClasses

                def controllerClass = applicationControllerClasses.find() { it.logicalPropertyName == controllerName }
                if (!controllerClass) {
                    log.warn("Couldn't find a Controller class associated with '${controllerName}'")
                    return true;
                }
                else log.debug("Controller class '${controllerClass}' located")

                List requireAuthorization= []

                // Process Controller/Action supplied authenticate
                def defaultAuthorize= CH.config.crowdAuth.defaultAuthorize
                switch (defaultAuthorize) {
                case String:
                    requireAuthorization= [defaultAuthorize]
                    break
                case List:
                    requireAuthorization= defaultAuthorize
                    break
                case Map:
                    def all= defaultAuthorize['all']
                    if (all) {
                        switch (all) {
                        case String:
                            requireAuthorization.add(all)
                            break
                        case List:
                            requireAuthorization.addAll(all)
                        default:
                            log.warn("DefaultAuthorize was not correctly defined in configuration.")
                        }
                    }

                    def thisAction= defaultAuthorize[actionName? actionName : 'index']
                    if (thisAction) {
                        switch (thisAction) {
                        case String:
                            requireAuthorization.add(thisAction)
                            break
                        case List:
                            requireAuthorization.addAll(thisAction)
                            break
                        default:
                            log.warn("DefaultAuthorize was not correctly defined in configuration.")
                        }
                    }
                }

                // Process Controller/Action supplied authorize
                Object authorize= GrailsClassUtils.getStaticPropertyValue(controllerClass.getClazz(), "authorize");
                switch (authorize) {
                case String:
                    requireAuthorization= [authorize]
                    break
                case List:
                    requireAuthorization= authorize
                    break
                case Map:
                    def all= authorize['all']
                    if (all) {
                        switch (all) {
                        case String:
                            requireAuthorization.add(all)
                            break
                        case List:
                            requireAuthorization.addAll(all)
                        default:
                            log.warn("Authorize was not correctly defined in Controller '${controllerClass}'.")
                        }
                    }

                    def thisAction= authorize[actionName? actionName : 'index']
                    if (thisAction) {
                        switch (thisAction) {
                        case String:
                            requireAuthorization.add(thisAction)
                            break
                        case List:
                            requireAuthorization.addAll(thisAction)
                            break
                        default:
                            log.warn("Authorize was not correctly defined in Controller '${controllerClass}'.")
                        }
                    }
                }

                // Process all authorization Groups
                log.debug("requireAuthorization: ${requireAuthorization}")
                if (requireAuthorization) {
                    if (!session.principalGroups?.containsAll(requireAuthorization)) {
                        response.sendError(403)
                    }
                }
            }
        }
    }
}

